@@ -498,34 +498,25 @@ static int alloc_list(struct net_device *dev)
 	for (i = 0; i < RX_RING_SIZE; i++) {
 		/* Allocated fixed size of skbuff */
 		struct sk_buff *skb;
-		dma_addr_t addr;
 
 		skb = netdev_alloc_skb_ip_align(dev, np->rx_buf_sz);
 		np->rx_skbuff[i] = skb;
-		if (!skb)
-			goto err_free_list;
-
-		addr = dma_map_single(&np->pdev->dev, skb->data,
-				      np->rx_buf_sz, DMA_FROM_DEVICE);
-		if (dma_mapping_error(&np->pdev->dev, addr))
-			goto err_kfree_skb;
+		if (!skb) {
+			free_list(dev);
+			return -ENOMEM;
+		}
 
 		np->rx_ring[i].next_desc = cpu_to_le64(np->rx_ring_dma +
 						((i + 1) % RX_RING_SIZE) *
 						sizeof(struct netdev_desc));
 		/* Rubicon now supports 40 bits of addressing space. */
-		np->rx_ring[i].fraginfo = cpu_to_le64(addr);
+		np->rx_ring[i].fraginfo =
+		    cpu_to_le64(dma_map_single(&np->pdev->dev, skb->data,
+					       np->rx_buf_sz, DMA_FROM_DEVICE));
 		np->rx_ring[i].fraginfo |= cpu_to_le64((u64)np->rx_buf_sz << 48);
 	}
 
 	return 0;
-
-err_kfree_skb:
-	dev_kfree_skb(np->rx_skbuff[i]);
-	np->rx_skbuff[i] = NULL;
-err_free_list:
-	free_list(dev);
-	return -ENOMEM;
 }
 
 static void rio_hw_init(struct net_device *dev)
@@ -962,18 +953,15 @@ receive_packet (struct net_device *dev)
 		} else {
 			struct sk_buff *skb;
 
-			skb = NULL;
 			/* Small skbuffs for short packets */
-			if (pkt_len <= copy_thresh)
-				skb = netdev_alloc_skb_ip_align(dev, pkt_len);
-			if (!skb) {
+			if (pkt_len > copy_thresh) {
 				dma_unmap_single(&np->pdev->dev,
 						 desc_to_dma(desc),
 						 np->rx_buf_sz,
 						 DMA_FROM_DEVICE);
 				skb_put (skb = np->rx_skbuff[entry], pkt_len);
 				np->rx_skbuff[entry] = NULL;
-			} else {
+			} else if ((skb = netdev_alloc_skb_ip_align(dev, pkt_len))) {
 				dma_sync_single_for_cpu(&np->pdev->dev,
 							desc_to_dma(desc),
 							np->rx_buf_sz,
@@ -1103,7 +1091,7 @@ get_stats (struct net_device *dev)
 	dev->stats.rx_bytes += dr32(OctetRcvOk);
 	dev->stats.tx_bytes += dr32(OctetXmtOk);
 
-	dev->stats.multicast += dr32(McstFramesRcvdOk);
+	dev->stats.multicast = dr32(McstFramesRcvdOk);
 	dev->stats.collisions += dr32(SingleColFrames)
 			     +  dr32(MultiColFrames);
 
@@ -1802,7 +1790,7 @@ rio_close (struct net_device *dev)
 	rio_hw_stop(dev);
 
 	free_irq(pdev->irq, dev);
-	del_timer_sync (&np->timer);
+	timer_delete_sync(&np->timer);
 
 	free_list(dev);
 
@@ -1842,7 +1830,7 @@ static int rio_suspend(struct device *device)
 		return 0;
 
 	netif_device_detach(dev);
-	del_timer_sync(&np->timer);
+	timer_delete_sync(&np->timer);
 	rio_hw_stop(dev);
 
 	return 0;
